@@ -1,5 +1,7 @@
 /*
  * Copyright (C) 2022 Purism SPC
+ *               2023-2024 The Phosh Developers
+ *               2025 Phosh.mobi e.V.
  *
  * SPDX-License-Identifier: GPL-3.0-or-later
  *
@@ -13,9 +15,6 @@
 #include "pos-completer-base.h"
 #include "pos-completer-priv.h"
 #include "pos-completer-presage.h"
-#include "pos-emoji-db.h"
-
-#include "util.h"
 
 #include <presage.h>
 
@@ -43,8 +42,6 @@ enum {
   PROP_0,
   PROP_NAME,
   PROP_PREEDIT,
-  PROP_BEFORE_TEXT,
-  PROP_AFTER_TEXT,
   PROP_COMPLETIONS,
   PROP_DICT_DIR,
   PROP_LAST_PROP
@@ -62,15 +59,13 @@ struct _PosCompleterPresage {
   PosCompleterBase      parent;
 
   char                 *name;
-  char                 *before_text;
-  char                 *after_text;
   GString              *preedit;
   GStrv                 completions;
+  GHashTable           *compmap; /* key: displayed completion, value: actual completion to use */
   guint                 max_completions;
 
   presage_t             presage;
   char                 *presage_past;
-  char                 *presage_future;
 
   char                 *dict_dir;
   char                 *lang;
@@ -91,11 +86,40 @@ G_DEFINE_TYPE_WITH_CODE (PosCompleterPresage, pos_completer_presage, POS_TYPE_CO
 
 
 static void
-pos_completer_presage_set_completions (PosCompleter *iface,
-                                       GStrv         completions,
-                                       gboolean      additional_sources)
+pos_completer_presage_update_compmap (PosCompleterPresage *self,
+                                      GStrv                completions,
+                                      GStrv                caps_completions,
+                                      GHashTable          *compmap)
+{
+  g_hash_table_remove_all (self->compmap);
+
+  if (g_strv_length (completions) != g_strv_length (caps_completions)) {
+    g_warning ("Capitalized completions don't match: %d != %d",
+               g_strv_length (completions),
+               g_strv_length (caps_completions));
+    return;
+  }
+
+  if (!compmap)
+    return;
+
+  /* Transform the keys from lowercase to their capitalized versions */
+  for (int i = 0; completions[i]; i++) {
+    gpointer value;
+
+    /* No need to free key as passed in `compmap` doesn't have key ownership */
+    if (g_hash_table_steal_extended (compmap, completions[i], NULL, &value))
+      g_hash_table_insert (self->compmap, g_strdup (caps_completions[i]), value);
+  }
+}
+
+
+static void
+pos_completer_presage_set_completions (PosCompleterPresage *self,
+                                       GStrv                completions,
+                                       gboolean             additional_sources,
+                                       GHashTable          *compmap)
 {
-  PosCompleterPresage *self = POS_COMPLETER_PRESAGE (iface);
   g_auto (GStrv) additional_results = NULL;
   g_auto (GStrv) caps_completions = NULL;
   g_autoptr (GStrvBuilder) builder = g_strv_builder_new ();
@@ -106,8 +130,11 @@ pos_completer_presage_set_completions (PosCompleter *iface,
                                                                     MAX_ADDITIONAL_RESULTS);
 
   caps_completions = pos_completer_capitalize_by_template (self->preedit->str, completions);
-  if (caps_completions)
+
+  if (caps_completions) {
     g_strv_builder_addv (builder, (const char **)caps_completions);
+    pos_completer_presage_update_compmap (self, completions, caps_completions, compmap);
+  }
   if (additional_results)
     g_strv_builder_addv (builder, (const char **)additional_results);
   g_strfreev (self->completions);
@@ -121,15 +148,32 @@ static void
 pos_completer_presage_predict (PosCompleterPresage *self)
 {
   presage_error_code_t result;
-  g_auto (GStrv) completions = NULL;
+  g_auto (GStrv) predictions = NULL;
 
-  result = presage_predict (self->presage, &completions);
+  result = presage_predict (self->presage, &predictions);
 
   if (result == PRESAGE_OK) {
-    pos_completer_presage_set_completions (POS_COMPLETER (self), completions, TRUE);
+    g_autoptr (GStrvBuilder) builder = g_strv_builder_new ();
+    g_auto (GStrv) completions = NULL;
+    g_autoptr (GHashTable) compmap = NULL;
+
+    compmap = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
+    for (int i = 0; predictions[i]; i++) {
+      g_autofree char *str = NULL;
+      g_autofree char *compl = NULL;
+
+      result = presage_completion (self->presage, predictions[i], &str);
+      if (result != PRESAGE_OK)
+        continue;
+
+      /* preedit is part of `presage_past` so we need to add it to the completion */
+      compl = g_strdup_printf ("%s%s", self->preedit->str, str);
+      g_hash_table_insert (compmap, predictions[i], g_steal_pointer (&compl));
+    }
+    pos_completer_presage_set_completions (self, predictions, TRUE, compmap);
   } else {
     g_warning ("Failed to complete %s", self->preedit->str);
-    pos_completer_presage_set_completions (POS_COMPLETER (self), NULL, FALSE);
+    pos_completer_presage_set_completions (self, NULL, FALSE, NULL);
   }
 }
 
@@ -156,31 +200,13 @@ pos_completer_presage_set_preedit (PosCompleter *iface, const char *preedit)
     g_string_append (self->preedit, preedit);
   else {
     /* No string: reset completions */
-    pos_completer_presage_set_completions (POS_COMPLETER (self), NULL, FALSE);
+    pos_completer_presage_set_completions (self, NULL, FALSE, NULL);
   }
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_PREEDIT]);
 }
 
 
-static const char *
-pos_completer_presage_get_before_text (PosCompleter *iface)
-{
-  PosCompleterPresage *self = POS_COMPLETER_PRESAGE (iface);
-
-  return self->before_text;
-}
-
-
-static const char *
-pos_completer_presage_get_after_text (PosCompleter *iface)
-{
-  PosCompleterPresage *self = POS_COMPLETER_PRESAGE (iface);
-
-  return self->after_text;
-}
-
-
 static void
 pos_completer_presage_set_surrounding_text (PosCompleter *iface,
                                             const char   *before_text,
@@ -188,22 +214,11 @@ pos_completer_presage_set_surrounding_text (PosCompleter *iface,
 {
   PosCompleterPresage *self = POS_COMPLETER_PRESAGE (iface);
 
-  if (g_strcmp0 (self->after_text, after_text) == 0 &&
-      g_strcmp0 (self->before_text, before_text) == 0) {
-    return;
-  }
-
-  g_free (self->after_text);
-  self->after_text = g_strdup (after_text);
-
-  g_free (self->before_text);
-  self->before_text = g_strdup (before_text);
+  pos_completer_base_set_surrounding_text (POS_COMPLETER_BASE (self),
+                                           before_text,
+                                           after_text);
 
   pos_completer_presage_predict (self);
-
-  g_debug ("Updating:  b:'%s', a:'%s'", self->before_text, self->after_text);
-  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_BEFORE_TEXT]);
-  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_AFTER_TEXT]);
 }
 
 
@@ -299,6 +314,17 @@ pos_completer_presage_set_language (PosCompleter *completer,
 }
 
 
+static const char *
+pos_completer_presage_lookup_completion (PosCompleter *completer, const char *completion)
+{
+  PosCompleterPresage *self = POS_COMPLETER_PRESAGE (completer);
+  const char *lookup;
+
+  lookup = g_hash_table_lookup (self->compmap, completion);
+  return lookup ? lookup : completion;
+}
+
+
 static void
 pos_completer_presage_set_property (GObject      *object,
                                     guint         property_id,
@@ -336,12 +362,6 @@ pos_completer_presage_get_property (GObject    *object,
   case PROP_PREEDIT:
     g_value_set_string (value, self->preedit->str);
     break;
-  case PROP_BEFORE_TEXT:
-    g_value_set_string (value, self->before_text);
-    break;
-  case PROP_AFTER_TEXT:
-    g_value_set_string (value, self->after_text);
-    break;
   case PROP_COMPLETIONS:
     g_value_set_boxed (value, self->completions);
     break;
@@ -358,11 +378,9 @@ pos_completer_presage_finalize (GObject *object)
   PosCompleterPresage *self = POS_COMPLETER_PRESAGE (object);
 
   g_clear_pointer (&self->completions, g_strfreev);
+  g_clear_pointer (&self->compmap, g_hash_table_unref);
   g_string_free (self->preedit, TRUE);
-  g_clear_pointer (&self->before_text, g_free);
-  g_clear_pointer (&self->after_text, g_free);
   g_clear_pointer (&self->presage_past, g_free);
-  g_clear_pointer (&self->presage_future, g_free);
   g_clear_pointer (&self->lang, g_free);
   g_clear_pointer (&self->dict_dir, g_free);
   presage_free (self->presage);
@@ -386,12 +404,6 @@ pos_completer_presage_class_init (PosCompleterPresageClass *klass)
   g_object_class_override_property (object_class, PROP_PREEDIT, "preedit");
   props[PROP_PREEDIT] = g_object_class_find_property (object_class, "preedit");
 
-  g_object_class_override_property (object_class, PROP_BEFORE_TEXT, "before-text");
-  props[PROP_BEFORE_TEXT] = g_object_class_find_property (object_class, "before-text");
-
-  g_object_class_override_property (object_class, PROP_AFTER_TEXT, "after-text");
-  props[PROP_AFTER_TEXT] = g_object_class_find_property (object_class, "after-text");
-
   g_object_class_override_property (object_class, PROP_COMPLETIONS, "completions");
   props[PROP_COMPLETIONS] = g_object_class_find_property (object_class, "completions");
 
@@ -407,9 +419,12 @@ static const char*
 pos_completer_presage_get_past_stream (void *data)
 {
   PosCompleterPresage *self = POS_COMPLETER_PRESAGE (data);
+  PosCompleterBase *base = POS_COMPLETER_BASE (self);
 
   g_free (self->presage_past);
-  self->presage_past = g_strdup_printf ("%s%s", self->before_text ?: "", self->preedit->str);
+  self->presage_past = g_strdup_printf ("%s%s",
+                                        pos_completer_base_get_before_text (base),
+                                        self->preedit->str);
 
   g_debug ("Past: %s", self->presage_past);
   return self->presage_past;
@@ -491,9 +506,17 @@ pos_completer_presage_feed_symbol (PosCompleter *iface, const char *symbol)
   g_autofree char *preedit = g_strdup (self->preedit->str);
 
   if (pos_completer_add_preedit (POS_COMPLETER (self), self->preedit, symbol)) {
-    self->updating_preedit = TRUE;
+    PosCompleterBase *base = POS_COMPLETER_BASE (self);
+    int before = 0;
 
-    g_signal_emit_by_name (self, "commit-string", self->preedit->str);
+    self->updating_preedit = TRUE;
+    if (pos_completer_base_wants_punctuation_swap (base, symbol) &&
+        /* Only swap if preedit is symbol + space */
+        self->preedit->len == 2) {
+      before = 1;
+    }
+
+    g_signal_emit_by_name (self, "commit-string", self->preedit->str, before, 0);
     pos_completer_presage_set_preedit (POS_COMPLETER (self), NULL);
     /* Make sure enter is processed as raw keystroke */
     if (g_strcmp0 (symbol, "KEY_ENTER") == 0) {
@@ -525,10 +548,9 @@ pos_completer_presage_interface_init (PosCompleterInterface *iface)
   iface->feed_symbol = pos_completer_presage_feed_symbol;
   iface->get_preedit = pos_completer_presage_get_preedit;
   iface->set_preedit = pos_completer_presage_set_preedit;
-  iface->get_before_text = pos_completer_presage_get_before_text;
-  iface->get_after_text = pos_completer_presage_get_after_text;
   iface->set_surrounding_text = pos_completer_presage_set_surrounding_text;
   iface->set_language = pos_completer_presage_set_language;
+  iface->lookup_completion = pos_completer_presage_lookup_completion;
 }
 
 
@@ -538,8 +560,7 @@ pos_completer_presage_init (PosCompleterPresage *self)
   self->max_completions = MAX_COMPLETIONS;
   self->preedit = g_string_new (NULL);
   self->name = "presage";
-  self->before_text = g_strdup ("");
-  self->after_text = g_strdup ("");
+  self->compmap = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 }
 
 /**
